{
  "nbformat": 4,
  "nbformat_minor": 0,
  "metadata": {
    "colab": {
      "provenance": []
    },
    "kernelspec": {
      "name": "python3",
      "display_name": "Python 3"
    },
    "language_info": {
      "name": "python"
    }
  },
  "cells": [
    {
      "cell_type": "markdown",
      "source": [
        "# オブジェクト指向回\n",
        "リファクタリング後のコード"
      ],
      "metadata": {
        "id": "0MfHnfKUKLUF"
      }
    },
    {
      "cell_type": "markdown",
      "source": [
        "## キャンペーンと継承の話"
      ],
      "metadata": {
        "id": "57mXj-ktJrjr"
      }
    },
    {
      "cell_type": "code",
      "source": [
        "from typing import final\n",
        "from typing_extensions import override"
      ],
      "metadata": {
        "id": "09_STXSzKxsZ"
      },
      "execution_count": null,
      "outputs": []
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "id": "VzO_ni5-JOex"
      },
      "outputs": [],
      "source": [
        "class Campaign:\n",
        "    \"キャンペーン\"\n",
        "\n",
        "    def get_discounted_price(self, price: int):\n",
        "        \"割引後の価格を計算する\"\n",
        "        raise NotImplementedError()\n",
        "\n",
        "    @final\n",
        "    def compute_discounted_amount(self, price: int):\n",
        "        \"割引された額を計算する\"\n",
        "        return price - self.get_discounted_price(price)\n",
        "\n",
        "\n",
        "class RateDiscountCampaign(Campaign):\n",
        "    \"割引キャンペーン\"\n",
        "\n",
        "    def __init__(self, discount_rate: float):\n",
        "        super().__init__()\n",
        "\n",
        "        self.__discount_rate = discount_rate\n",
        "\n",
        "    @override\n",
        "    def get_discounted_price(self, price: int):\n",
        "        return int(price * (1 - self.__discount_rate))\n",
        "\n",
        "\n",
        "class ValueDiscountCampaign(Campaign):\n",
        "    \"固定値値引きキャンペーン\"\n",
        "\n",
        "    def __init__(self, discount_value: float):\n",
        "        super().__init__()\n",
        "\n",
        "        self.__discount_value = discount_value\n",
        "\n",
        "    @override\n",
        "    def get_discounted_price(self, price: int):\n",
        "        return price - self.__discount_value\n",
        "\n",
        "\n",
        "class Subscription:\n",
        "    \"申込\"\n",
        "\n",
        "    def __init__(self, plan_id: int, price: int, campaign: Campaign):\n",
        "        self.__plan_id = plan_id\n",
        "        self.__monthly_price = price\n",
        "        self.__campaign = campaign\n",
        "\n",
        "    def plan_id(self):\n",
        "        return self.__plan_id\n",
        "\n",
        "    def monthly_price(self):\n",
        "        \"割引後の価格を計算する\"\n",
        "        return self.__campaign.get_discounted_price(self.__monthly_price)\n",
        "\n",
        "    def compute_discounted_amount(self):\n",
        "        \"割引される額を計算する\"\n",
        "        return self.__campaign.compute_discounted_amount(self.__monthly_price)\n",
        "\n",
        "\n"
      ]
    },
    {
      "cell_type": "code",
      "source": [
        "sub = Subscription(\n",
        "    plan_id=12345,\n",
        "    price=1000,  # 1000円のプランを購読\n",
        "    campaign=RateDiscountCampaign(0.2),  # 20% OFF\n",
        ")\n",
        "print(sub.monthly_price())  # => 800円\n",
        "print(sub.compute_discounted_amount())  # => 200円\n",
        "\n",
        "sub = Subscription(\n",
        "    plan_id=12345,\n",
        "    price=1000,  # 1000円のプランを購読\n",
        "    campaign=ValueDiscountCampaign(300),  # 300円引き\n",
        ")\n",
        "print(sub.monthly_price())  # => 700円\n",
        "print(sub.compute_discounted_amount())  # => 300円\n",
        "\n"
      ],
      "metadata": {
        "id": "jpsr4D4vJi_H",
        "colab": {
          "base_uri": "https://localhost:8080/"
        },
        "outputId": "4c1ce10c-da62-4674-f5d3-e56cbe7fb41e"
      },
      "execution_count": null,
      "outputs": [
        {
          "output_type": "stream",
          "name": "stdout",
          "text": [
            "800\n",
            "200\n",
            "700\n",
            "300\n"
          ]
        }
      ]
    },
    {
      "cell_type": "markdown",
      "source": [
        "## メール送信とインターフェースの話"
      ],
      "metadata": {
        "id": "qaS2GmoFJ7eq"
      }
    },
    {
      "cell_type": "code",
      "source": [
        "from abc import ABC, abstractmethod\n"
      ],
      "metadata": {
        "id": "2Th2qUB2J-HL"
      },
      "execution_count": null,
      "outputs": []
    },
    {
      "cell_type": "code",
      "source": [
        "def get_price(plan_id: int) -> int:\n",
        "    # 本当は価格を取得する処理\n",
        "    return 1000\n",
        "\n",
        "\n",
        "def get_campaign() -> Campaign:\n",
        "    return RateDiscountCampaign(0.2)\n",
        "\n",
        "\n",
        "def save_to_database(subscription: Subscription):\n",
        "    # 本当は保存処理\n",
        "    pass\n",
        "\n",
        "\n",
        "class MailClient(ABC):\n",
        "    @abstractmethod\n",
        "    def send_mail_to(self, address: str, message: str):\n",
        "        \"メール送信処理\"\n",
        "        pass\n",
        "\n",
        "\n",
        "# 本番環境用\n",
        "class ActualMailClient(MailClient):\n",
        "    @override\n",
        "    def send_mail_to(self, address: str, message: str):\n",
        "        # 本当はメール送信処理、例示のため print() で代用しておく\n",
        "        print(\"actually send mail\")\n",
        "        pass\n",
        "\n",
        "\n",
        "def subscribe(mail_client: MailClient, mail_address: str, plan_id: int):\n",
        "    \"申し込む\"\n",
        "\n",
        "    # なんやかんやプラン情報を取得して\n",
        "    price = get_price(plan_id)\n",
        "    campaign = get_campaign()\n",
        "    subscription = Subscription(plan_id, price, campaign=campaign)\n",
        "\n",
        "    # 申込情報をデータベースに保存\n",
        "    save_to_database(subscription)\n",
        "\n",
        "    # メール送信\n",
        "    message = f\"{subscription.monthly_price()}円で購読ありがとうございます\"\n",
        "    mail_client.send_mail_to(mail_address, message)\n",
        "\n",
        "\n"
      ],
      "metadata": {
        "id": "hVQCzpXEJ3iZ"
      },
      "execution_count": null,
      "outputs": []
    },
    {
      "cell_type": "code",
      "source": [
        "# 本番\n",
        "subscribe(ActualMailClient(), \"test@example.com\", 12345)\n"
      ],
      "metadata": {
        "id": "CxBybVxlJ3kJ",
        "colab": {
          "base_uri": "https://localhost:8080/"
        },
        "outputId": "c0ee9cc6-54a0-4e7d-cad6-cb0fc83b6da3"
      },
      "execution_count": null,
      "outputs": [
        {
          "output_type": "stream",
          "name": "stdout",
          "text": [
            "actually send mail\n"
          ]
        }
      ]
    },
    {
      "cell_type": "markdown",
      "source": [
        "### テスト用"
      ],
      "metadata": {
        "id": "lrtuF50wKXek"
      }
    },
    {
      "cell_type": "code",
      "source": [
        "class MockMailClient(MailClient):\n",
        "    @override\n",
        "    def send_mail_to(self, address: str, message: str):\n",
        "        # テスト用なので、メール送信をしない\n",
        "        print(\"送信しました\")\n"
      ],
      "metadata": {
        "id": "46a4yNB5J3mT"
      },
      "execution_count": null,
      "outputs": []
    },
    {
      "cell_type": "code",
      "source": [
        "subscribe(MockMailClient(), \"test@example.com\", 12345)"
      ],
      "metadata": {
        "id": "ldXWgAnAJ3ot",
        "colab": {
          "base_uri": "https://localhost:8080/"
        },
        "outputId": "cacc3763-2506-4768-b88f-83fb4e24a56c"
      },
      "execution_count": null,
      "outputs": [
        {
          "output_type": "stream",
          "name": "stdout",
          "text": [
            "送信しました\n"
          ]
        }
      ]
    },
    {
      "cell_type": "code",
      "source": [],
      "metadata": {
        "id": "52u5F3tQKaak"
      },
      "execution_count": null,
      "outputs": []
    },
    {
      "cell_type": "markdown",
      "source": [
        "  "
      ],
      "metadata": {
        "id": "4nURCuq5KbGF"
      }
    }
  ]
}